@@ -16,8 +16,10 @@
 -- You should have received a copy of the GNU Affero General Public License
 -- along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+{-# LANGUAGE DeriveGeneric #-}
+
 import HOpenPGP.Tools.Common (banner, versioner, warranty, keyMatchesFingerprint, keyMatchesEightOctetKeyId, keyMatchesUIDSubString)
-import HOpenPGP.Tools.ExpressionParsing (pPE)
+import HOpenPGP.Tools.Parser (parseTKExp)
 import Codec.Encryption.OpenPGP.Fingerprint (fingerprint, eightOctetKeyID)
 import Codec.Encryption.OpenPGP.KeyInfo (pubkeySize, pkalgoAbbrev)
 import Codec.Encryption.OpenPGP.KeySelection (parseEightOctetKeyId, parseFingerprint)
@@ -29,49 +31,56 @@ import Control.Arrow ((&&&))
 import Control.Lens ((^.), _1, _2, (^..))
 import Control.Monad.Trans.Resource (runResourceT, MonadResource)
 import qualified Control.Monad.Trans.State.Lazy as S
-import Control.Monad.Trans.Writer.Lazy (execWriter, tell)
-import qualified Data.Attoparsec.Text as A
+import qualified Data.Aeson as A
 import qualified Data.ByteString as B
+import qualified Data.ByteString.Lazy as BL
 import Data.Conduit (($=),($$), Source)
 import qualified Data.Conduit.Binary as CB
 import Data.Conduit.Cereal (conduitGet)
 import qualified Data.Conduit.List as CL
-import Data.Conduit.OpenPGP.Filter (Expr(..), PKPPredicate(..), PKPOp(..), PKPVar(..), PKPValue(..))
+import Data.Conduit.OpenPGP.Filter (conduitTKFilter, FilterPredicates(RTKFilterPredicate))
 import Data.Conduit.OpenPGP.Keyring (conduitToTKsDropping, sinkKeyringMap)
 import Data.Data.Lens (biplate)
 import Data.Either (rights)
 import qualified Data.IxSet as IxSet
-import Data.Graph.Inductive.Graph (Graph(mkGraph), emap)
+import Data.Graph.Inductive.Graph (Graph(mkGraph), emap, Path, prettyPrint)
 import Data.Graph.Inductive.PatriciaTree (Gr)
 import Data.Graph.Inductive.Query.SP (sp)
-import Data.GraphViz (graphToDot, nonClusteredParams)
+import Data.GraphViz (graphToDot, nonClusteredParams, GraphvizParams(..))
+import Data.GraphViz.Attributes (toLabel)
 import Data.GraphViz.Types (printDotGraph)
 import Data.HashMap.Lazy (HashMap)
 import qualified Data.HashMap.Lazy as HashMap
 import Data.List (nub, sort)
+import Data.Map (Map)
+import qualified Data.Map as Map
 import Data.Maybe (fromMaybe, mapMaybe, listToMaybe)
-import Data.Monoid ((<>))
+import Data.Monoid ((<>), mconcat)
 import Data.Serialize (get, put, runPut)
 import qualified Data.Text as T
 import qualified Data.Text.Lazy.IO as TLIO
 import Data.Time.Clock.POSIX (getPOSIXTime, posixSecondsToUTCTime)
 import Data.Traversable (traverse)
 import Data.Tuple (swap)
+import qualified Data.Yaml as Y
+import GHC.Generics
 import System.Directory (getHomeDirectory)
 
-import Options.Applicative.Builder (argument, command, footer, header, help, info, long, metavar, option, prefs, progDesc, showDefault, showHelpOnError, str, strOption, subparser, switch, value)
+import Options.Applicative.Builder (argument, auto, command, footerDoc, headerDoc, help, helpDoc, info, long, metavar, option, prefs, progDesc, showDefault, showHelpOnError, str, strOption, subparser, switch, value)
 import Options.Applicative.Extra (customExecParser, helper)
 import Options.Applicative.Types (Parser)
 
-import System.IO (Handle, hFlush, hPutStrLn, stderr, hSetBuffering, BufferMode(..))
+import System.IO (Handle, hFlush, stderr, hSetBuffering, BufferMode(..))
+import Text.PrettyPrint.ANSI.Leijen ((<+>), (</>), hardline, hPutDoc, list, putDoc, text)
 
 grabMatchingKeysConduit :: MonadResource m => FilePath -> Bool -> String -> Source m TK
-grabMatchingKeysConduit fp filt srch = CB.sourceFile fp $= conduitGet get $= conduitToTKsDropping $= CL.filter (if filt then filterMatch else matchAny)
+grabMatchingKeysConduit fp filt srch = CB.sourceFile fp $= conduitGet get $= conduitToTKsDropping $= (if filt then conduitTKFilter ufp else CL.filter matchAny)
     where
         matchAny tk = either (const False) id $ fmap (keyMatchesFingerprint True tk) efp <|> fmap (keyMatchesEightOctetKeyId True tk . Right) eeok <|> return (keyMatchesUIDSubString srch tk)
-        filterMatch tk = eval pkpEval (either error id (A.parseOnly pPE (T.pack srch))) (tk^.tkKey._1)
         efp = parseFingerprint . T.pack $ srch
         eeok = parseEightOctetKeyId . T.pack $ srch
+        ufp = RTKFilterPredicate (parseE srch)
+        parseE e = either (error . ("filter parse error: "++)) id (parseTKExp e) -- this should be more specialized
 
 grabMatchingKeys :: FilePath -> Bool -> String -> IO [TK]
 grabMatchingKeys fp filt srch = runResourceT $ grabMatchingKeysConduit fp filt srch $$ CL.consume
@@ -79,15 +88,49 @@ grabMatchingKeys fp filt srch = runResourceT $ grabMatchingKeysConduit fp filt s
 grabMatchingKeysKeyring :: FilePath -> Bool -> String -> IO Keyring
 grabMatchingKeysKeyring fp filt srch = runResourceT $ grabMatchingKeysConduit fp filt srch $$ sinkKeyringMap
 
-showKey :: TK -> IO ()
-showKey key = putStrLn . unlines . execWriter $ do
-    tell [ "pub   " ++ either (const "unknown") show (pubkeySize (key^.tkKey._1.pubkey)) ++ pkalgoAbbrev (key^.tkKey._1.pkalgo) ++ "/0x" ++ (either (const "unknown") show . eightOctetKeyID $ key^.tkKey._1 ) ]
-    tell $ map (\(x,_) -> "uid                            " ++ x) (key^.tkUIDs)
-    tell $ map (\(PublicSubkeyPkt x,_) -> "sub   " ++ either (const "unknown") show (pubkeySize (x^.pubkey)) ++ pkalgoAbbrev (x^.pkalgo) ++ "/0x" ++ (either (const "unknown") show . eightOctetKeyID $ x)) (key^.tkSubs)
+data Key = Key {
+    keysize :: Maybe Int
+  , keyalgo :: String
+  , keyalgoabbreviation :: String
+  , keyid :: String
+  , fpr :: String
+} deriving Generic
+
+data TKey = TKey {
+    publickey :: Key
+  , uids :: [String]
+  , subkeys :: [Key]
+} deriving Generic
+
+instance A.ToJSON Key
+instance A.ToJSON TKey
+
+tkToTKey :: TK -> TKey
+tkToTKey tk = TKey {
+    publickey = mkey (tk^.tkKey._1)
+  , uids = tk^.tkUIDs^..traverse._1
+  , subkeys = map (mkey . \(PublicSubkeyPkt x,_) -> x) (tk^.tkSubs)
+}
+    where
+        mkey = Key <$> either (const Nothing) Just . pubkeySize . _pubkey
+                   <*> show . _pkalgo
+                   <*> pkalgoAbbrev . _pkalgo
+                   <*> either (const "unknown") show . eightOctetKeyID
+                   <*> show . fingerprint
+
+showTKey :: TKey -> IO ()
+showTKey key = putDoc $
+    text "pub  " <+> sizeabbrevkeyid (publickey key) <> hardline <>
+    mconcat (map (\x -> text "uid                           " <+> text x <> hardline) (uids key)) <>
+    mconcat (map (\x -> text "sub  " <+> sizeabbrevkeyid x <> hardline) (subkeys key)) <>
+    hardline
+    where
+        sizeabbrevkeyid k = text (maybe "unknown" show (keysize k)) <> text (keyalgoabbreviation k) <> text "/0x" <> text (keyid k)
 
 data Options = Options {
     keyring :: String
   , graphOutputFormat :: GraphOutputFormat
+  , pathsOutputFormat :: PathsOutputFormat
   , targetIsFilter :: Bool
   , target1 :: String
   , target2 :: String
@@ -96,7 +139,10 @@ data Options = Options {
 
 data Command = CmdList Options | CmdExportPubkeys Options | CmdGraph Options | CmdFindPaths Options
 
-data GraphOutputFormat = GraphViz
+data GraphOutputFormat = GraphViz | LossyPretty
+    deriving (Bounded, Enum, Eq, Read, Show)
+
+data PathsOutputFormat = Unstructured | JSON | YAML
     deriving (Eq, Read, Show)
 
 listO :: String -> Parser Options
@@ -106,10 +152,18 @@ listO homedir = Options
        <> metavar "FILE"
        <> help "file containing keyring" )))
     <*> pure GraphViz -- unused
+    <*> option auto
+        ( long "output-format"
+       <> metavar "FORMAT"
+       <> value Unstructured
+       <> showDefault
+       <> help "output format" )
     <*> switch ( long "filter" <> help "treat target as filter" )
-    <*> argument str ( metavar "TARGET" )
+    <*> argument str ( metavar "TARGET" <> targetHelp )
     <*> pure ""
     <*> pure ""
+    where
+        targetHelp = helpDoc . Just $ text "target (which keys to output)*"
 
 graphO :: String -> Parser Options
 graphO homedir = Options
@@ -117,16 +171,21 @@ graphO homedir = Options
         ( long "keyring"
        <> metavar "FILE"
        <> help "file containing keyring" )))
-    <*> option
+    <*> option auto
         ( long "output-format"
        <> metavar "FORMAT"
        <> value GraphViz
        <> showDefault
-       <> help "output format" )
+       <> ofhelp )
+    <*> pure Unstructured -- unused
     <*> switch ( long "filter" <> help "treat target as filter" )
-    <*> argument str ( metavar "TARGET" )
+    <*> argument str ( metavar "TARGET" <> targetHelp )
     <*> pure ""
     <*> pure ""
+    where
+        ofhelp = helpDoc . Just $ text "output format" <> hardline <> list (map (text . show) ofchoices)
+        ofchoices = [minBound..maxBound] :: [GraphOutputFormat]
+        targetHelp = helpDoc . Just $ text "target (which keys to graph)*"
 
 findPathsO :: String -> Parser Options
 findPathsO homedir = Options
@@ -134,16 +193,21 @@ findPathsO homedir = Options
         ( long "keyring"
        <> metavar "FILE"
        <> help "file containing keyring" )))
-    <*> option
+    <*> pure GraphViz -- unused
+    <*> option auto
         ( long "output-format"
        <> metavar "FORMAT"
-       <> value GraphViz
+       <> value Unstructured
        <> showDefault
        <> help "output format" )
     <*> switch ( long "filter" <> help "treat targets as filter" )
-    <*> argument str ( metavar "TARGET-SET" )
-    <*> argument str ( metavar "FROM-KEYS" )
-    <*> argument str ( metavar "TO-KEYS" )
+    <*> argument str ( metavar "TARGET-SET" <> targetHelp )
+    <*> argument str ( metavar "FROM-KEYS" <> fromHelp )
+    <*> argument str ( metavar "TO-KEYS" <> toHelp )
+    where
+        targetHelp = helpDoc . Just $ text "target (which keys to use in pathfinding)*"
+        fromHelp = helpDoc . Just $ text "from (which keys to use for the source of paths)*"
+        toHelp = helpDoc . Just $ text "to (which keys to use for the destinations of paths)*"
 
 dispatch :: Command -> IO ()
 dispatch (CmdList o) = banner' stderr >> hFlush stderr >> doList o
@@ -155,22 +219,34 @@ main :: IO ()
 main = do
     hSetBuffering stderr LineBuffering
     homedir <- getHomeDirectory
-    customExecParser (prefs showHelpOnError) (info (helper <*> versioner <*> cmd homedir) (header (banner "hkt") <> progDesc "hOpenPGP Keyring Tool" <> footer (warranty "hkt"))) >>= dispatch
+    customExecParser (prefs showHelpOnError) (info (helper <*> versioner <*> cmd homedir) (headerDoc (Just (banner "hkt")) <> progDesc "hOpenPGP Keyring Tool" <> footerDoc (Just (warranty "hkt")))) >>= dispatch
 
 cmd :: String -> Parser Command
 cmd homedir = subparser
-    ( command "list" (info ( CmdList <$> listO homedir) ( progDesc "list matching keys" ))
-   <> command "export-pubkeys" (info ( CmdExportPubkeys <$> listO homedir) ( progDesc "export matching keys to stdout" ))
-   <> command "graph" (info ( CmdGraph <$> graphO homedir) ( progDesc "graph certifications" ))
-   <> command "findpaths" (info ( CmdFindPaths <$> findPathsO homedir) ( progDesc "find short paths between keys" )))
+    ( command "export-pubkeys" (info ( CmdExportPubkeys <$> listO homedir) ( progDesc "export matching keys to stdout" <> footerDoc (Just foot) ))
+   <> command "findpaths" (info ( CmdFindPaths <$> findPathsO homedir) ( progDesc "find short paths between keys" <> footerDoc (Just foot) ))
+   <> command "graph" (info ( CmdGraph <$> graphO homedir) ( progDesc "graph certifications" <> footerDoc (Just foot) ))
+   <> command "list" (info ( CmdList <$> listO homedir) ( progDesc "list matching keys" <> footerDoc (Just foot) ))
+    )
+    where
+        foot = hardline <> text "*if --filter is not specified, this must be"
+               </> text "a fingerprint," </> text "an eight-octet key ID,"
+               </> text "or a substring of a UID (including an empty string)" <> hardline
+               <> text "if --filter is specified, it must be"
+               </> text "something in filter syntax (see source)."
 
 banner' :: Handle -> IO ()
-banner' h = hPutStrLn h (banner "hkt" ++ "\n" ++ warranty "hkt")
+banner' h = hPutDoc h (banner "hkt" <> hardline <> warranty "hkt" <> hardline)
 
 doList :: Options -> IO ()
 doList o = do
-    keys <- grabMatchingKeys (keyring o) (targetIsFilter o) (target1 o)
-    mapM_ showKey keys
+    keys' <- grabMatchingKeys (keyring o) (targetIsFilter o) (target1 o)
+    let keys = map tkToTKey keys'
+    case pathsOutputFormat o of
+        Unstructured -> mapM_ showTKey keys
+        JSON -> BL.putStr . A.encode $ keys
+        YAML -> B.putStr . Y.encode $ keys
+    putStrLn ""
 
 doExportPubkeys :: Options -> IO ()
 doExportPubkeys o = do
@@ -191,7 +267,12 @@ doGraph :: Options -> IO ()
 doGraph o = do
     cpt <- getPOSIXTime
     kr <- grabMatchingKeysKeyring (keyring o) (targetIsFilter o) (target1 o)
-    TLIO.putStrLn . printDotGraph $ graphToDot nonClusteredParams (buildKeyGraph ((buildMaps &&& id) (rights (map (verifyTKWith (verifySigWith (verifyAgainstKeyring kr)) (Just (posixSecondsToUTCTime cpt))) (IxSet.toList kr)))))
+    let g = buildKeyGraph ((buildMaps &&& id) (rights (map (verifyTKWith (verifySigWith (verifyAgainstKeyring kr)) (Just (posixSecondsToUTCTime cpt))) (IxSet.toList kr))))
+    case graphOutputFormat o of
+        LossyPretty -> prettyPrint g
+        GraphViz -> TLIO.putStrLn . printDotGraph . graphToDot nonClusteredLabeledNodesParams $ g
+    where
+        nonClusteredLabeledNodesParams = nonClusteredParams { fmtNode = \(_,l) -> [toLabel $ show l] }
 
 buildMaps :: [TK] -> (KeyMaps, Int)
 buildMaps ks = S.execState (mapM_ mapsInsertions ks) (KeyMaps HashMap.empty HashMap.empty HashMap.empty, 0)
@@ -226,50 +307,46 @@ buildKeyGraph ((KeyMaps k2f f2i _, _), ks) = mkGraph nodes edges
         sigs tk = concat ((tk^..tkUIDs.traverse._2) ++ (tk^..tkUAts.traverse._2))
         samesies (x,y,_) = x == y
 
+data PaF = PaF {
+    certPaths :: [Path]
+  , keyFingerprints :: Map String TwentyOctetFingerprint
+} deriving Generic
+
+instance A.ToJSON PaF
+instance A.ToJSON TwentyOctetFingerprint where
+    toJSON = A.toJSON . show
+
 doFindPaths :: Options -> IO ()
 doFindPaths o = do
     cpt <- getPOSIXTime
     kr <- grabMatchingKeysKeyring (keyring o) (targetIsFilter o) (target1 o)
-    let keys1 = filter (if targetIsFilter o then filterMatch (target2 o) else matchAny (target2 o)) (IxSet.toList kr)
-        keys2 = filter (if targetIsFilter o then filterMatch (target3 o) else matchAny (target3 o)) (IxSet.toList kr)
-        ((KeyMaps k2f f2i i2f, i), ks) = (buildMaps &&& id) (rights (map (verifyTKWith (verifySigWith (verifyAgainstKeyring kr)) (Just (posixSecondsToUTCTime cpt))) (IxSet.toList kr)))
+    -- FIXME: clean this up
+    keys1 <- runResourceT $ CL.sourceList (IxSet.toList kr) $= (if filt then conduitTKFilter (ufpt (target2 o)) else CL.filter (matchAny (target2 o))) $$ CL.consume
+    keys2 <- runResourceT $ CL.sourceList (IxSet.toList kr) $= (if filt then conduitTKFilter (ufpt (target3 o)) else CL.filter (matchAny (target3 o))) $$ CL.consume
+    let ((KeyMaps k2f f2i i2f, i), ks) = (buildMaps &&& id) (rights (map (verifyTKWith (verifySigWith (verifyAgainstKeyring kr)) (Just (posixSecondsToUTCTime cpt))) (IxSet.toList kr)))
         keygraph = buildKeyGraph ((KeyMaps k2f f2i i2f, i), ks)
         keysToIs = mapMaybe (\x -> HashMap.lookup (fingerprint (x^.tkKey._1)) f2i)
         froms = keysToIs keys1
         tos = keysToIs keys2
         combos = froms >>= \f -> tos >>= \t -> return (f,t)
         paths = map (\(x,y) -> sp x y (emap (const (1.0 :: Double)) keygraph)) combos
-    print paths
-    putStrLn . unlines $ map (\x -> maybe (show x) show $ HashMap.lookup x i2f >>= \y -> return (x, y)) (nub (sort (concat paths)))
+        paf = PaF paths (Map.fromList (mapMaybe (\x -> HashMap.lookup x i2f >>= \y -> return (show x,y)) (nub (sort (concat paths)))))
+    case pathsOutputFormat o of
+        Unstructured -> do -- FIXME: do something about this
+                           putStrLn . unlines $ map (show . ((,) =<< length)) paths
+                           putStrLn . unlines $ map (\x -> maybe (show x) show $ HashMap.lookup x i2f >>= \y -> return (x, y)) (nub (sort (concat paths)))
+        JSON -> BL.putStr . A.encode $ paf
+        YAML -> B.putStr . Y.encode $ paf
+    putStrLn ""
     where  -- FIXME: deduplicate this
+        filt = targetIsFilter o
         matchAny srch tk = either (const False) id $ fmap (keyMatchesFingerprint True tk) (efp srch) <|> fmap (keyMatchesEightOctetKeyId True tk . Right) (eeok srch) <|> return (keyMatchesUIDSubString srch tk)
-        filterMatch srch tk = eval pkpEval (either error id (A.parseOnly pPE (T.pack srch))) (tk^.tkKey._1)
         efp srch = parseFingerprint . T.pack $ srch
         eeok srch = parseEightOctetKeyId . T.pack $ srch
+        ufpt srch = RTKFilterPredicate (parseE srch)
+        parseE e = either (error . ("filter parse error: "++)) id (parseTKExp e) -- this should be more specialized
 
 -- FIXME: deduplicate the following code
-eval :: (a -> v -> Bool) -> Expr a -> v -> Bool
-eval t e v = ev e
-  where
-        ev EAny = True
-        ev (EAnd e1 e2) = ev e1 && ev e2
-        ev (EOr e1 e2) =  ev e1 || ev e2
-        ev (ENot e1) = (not . ev) e1
-        ev (E e') = t e' v
-
-pkpEval :: PKPPredicate -> PKPayload -> Bool
-pkpEval (PKPPredicate lhs o rhs) pkp = uncurry (opreduce o) (vreduce (lhs,pkp),rhs)
-    where
-        opreduce PKEquals = (==)
-        opreduce PKLessThan = (<)
-        opreduce PKGreaterThan = (>)
-        vreduce (PKPVVersion, p) = PKPInt (kv (_keyVersion p))
-        vreduce (PKPVPKA, p) = PKPPKA (_pkalgo p)
-        vreduce (PKPVKeysize, p) = PKPInt (either (const 0) id . pubkeySize . _pubkey $ p) -- FIXME: this should be smarter
-        vreduce (PKPVTimestamp, p) = PKPInt (fromIntegral (_timestamp p))
-        kv DeprecatedV3 = 3
-        kv V4 = 4
-
 sigissuer :: SignaturePayload -> Maybe EightOctetKeyId
 getIssuer :: SigSubPacketPayload -> Maybe EightOctetKeyId
 hashAlgo :: SignaturePayload -> HashAlgorithm
@@ -284,4 +361,3 @@ getIssuer _ = Nothing
 
 hashAlgo (SigV4 _ _ x _ _ _ _) = x
 hashAlgo _ = error "V3 sig not supported here"
-
