@@ -22,12 +22,13 @@ import HOpenPGP.Tools.Common (banner, versioner, warranty)
 import Codec.Encryption.OpenPGP.Expirations (getKeyExpirationTimesFromSignature)
 import Codec.Encryption.OpenPGP.Fingerprint (fingerprint, eightOctetKeyID)
 import Codec.Encryption.OpenPGP.KeyInfo (pubkeySize, pkalgoAbbrev)
+import Codec.Encryption.OpenPGP.Serialize ()
 import Codec.Encryption.OpenPGP.Signatures (verifyTKWith, verifySigWith, verifyAgainstKeys)
 import Codec.Encryption.OpenPGP.Types
-import Control.Applicative ((<$>),(<*>))
+import Control.Applicative ((<$>),(<*>), pure)
 import Control.Arrow ((***), second)
 import Control.Error.Util (hush)
-import Control.Lens ((^.), (&), _1)
+import Control.Lens ((^.), (&), _1, _2, mapped, over)
 import Control.Monad.Trans.Resource (runResourceT)
 import Control.Monad.Trans.Writer.Lazy (execWriter, tell)
 import qualified Crypto.Hash.SHA3 as SHA3
@@ -38,28 +39,32 @@ import qualified Data.ByteString.Lazy as BL
 import qualified Data.ByteString.Base16 as Base16
 import Data.Conduit (($=),($$))
 import qualified Data.Conduit.Binary as CB
-import Data.Conduit.Cereal (conduitGet)
+import Data.Conduit.Cereal (conduitGet, conduitPut)
 import qualified Data.Conduit.List as CL
 import Data.Conduit.OpenPGP.Keyring (conduitToTKsDropping)
-import Data.List (unfoldr, elemIndex, findIndex, sortBy, intercalate)
+import Data.List (unfoldr, elemIndex, findIndex, nub, sort, sortBy, intercalate)
 import qualified Data.Map as Map
 import Data.Maybe (fromMaybe, mapMaybe, listToMaybe)
 import Data.Monoid ((<>), mconcat, Monoid(..))
 import Data.Ord (comparing, Down(..))
-import Data.Serialize (get)
+import Data.Serialize (get, put)
+import Data.Text (Text)
+import qualified Data.Text as T
+import Data.Text.Encoding (decodeUtf8With) -- FIXME: this should be made unnecessary by an API break in hOpenPGP
+import Data.Text.Encoding.Error (lenientDecode)
 import Data.Time.Clock.POSIX (getPOSIXTime, posixSecondsToUTCTime, POSIXTime)
 import Data.Time.Format (formatTime)
 import qualified Data.Yaml as Y
 import GHC.Generics
 
-import Options.Applicative.Builder (command, footer, header, help, info, long, metavar, prefs, progDesc, showDefault, showHelpOnError, option, subparser, value)
+import Options.Applicative.Builder (auto, command, footerDoc, headerDoc, helpDoc, info, long, metavar, prefs, progDesc, showDefault, showHelpOnError, option, subparser, value)
 import Options.Applicative.Extra (customExecParser, helper)
 import Options.Applicative.Types (Parser)
 
-import System.IO (Handle, hFlush, hPutStrLn, stderr, stdin, hSetBuffering, BufferMode(..))
+import System.IO (Handle, hFlush, stderr, stdin, stdout, hSetBuffering, BufferMode(..))
 import System.Locale (defaultTimeLocale)
 
-import Text.PrettyPrint.ANSI.Leijen (colon, green, indent, linebreak, list, putDoc, red, text, yellow, (<+>), Doc) -- need 0.6.7 for hardline
+import Text.PrettyPrint.ANSI.Leijen (colon, green, hardline, hPutDoc, indent, linebreak, list, putDoc, red, text, yellow, (<+>), Doc)
 
 data KAS = KAS {
     pubkeyalgo :: Colored PubKeyAlgorithm
@@ -84,7 +89,7 @@ data KeyReport = KeyReport {
   , keyVer :: Colored KeyVersion
   , keyCreationTime :: TimeStamp
   , keyAlgorithmAndSize :: KAS
-  , keyUIDsAndUAts :: FakeMap String UIDReport
+  , keyUIDsAndUAts :: FakeMap String (Colored UIDReport)
   , keyBestOf :: Maybe UIDReport
 } deriving Generic
 
@@ -92,6 +97,13 @@ data UIDReport = UIDReport {
     uidSelfSigHashAlgorithms :: [Colored HashAlgorithm]
   , uidPreferredHashAlgorithms :: [Colored [HashAlgorithm]]
   , uidKeyExpirationTimes :: [Colored [TimeStamp]]
+  , uidRevocationStatus :: [RevocationStatus]
+} deriving Generic
+
+data RevocationStatus = RevocationStatus {
+    isRevoked :: Bool
+  , revocationCode :: String
+  , revocationReason :: Text
 } deriving Generic
 
 instance A.ToJSON KAS
@@ -110,10 +122,11 @@ instance A.ToJSON KeyVersion where
     toJSON V4 = A.toJSON (4 :: Int)
 instance A.ToJSON TwentyOctetFingerprint where
     toJSON = A.toJSON . show
+instance A.ToJSON RevocationStatus
 
 instance Monoid UIDReport where
-    mempty = UIDReport [] [] []
-    mappend (UIDReport a b c) (UIDReport d e f) = UIDReport (a++d) (b++e) (c++f)
+    mempty = UIDReport [] [] [] []
+    mappend (UIDReport a b c d) (UIDReport a' b' c' d') = UIDReport (a++a') (b++b') (c++c') (d <> d')
 
 checkKey :: Maybe POSIXTime -> TK -> KeyReport
 checkKey mpt key = (\x -> x { keyBestOf = populateBestOf x }) KeyReport {
@@ -126,7 +139,7 @@ checkKey mpt key = (\x -> x { keyBestOf = populateBestOf x }) KeyReport {
     , pubkeysize = key^.tkKey._1.pubkey & colorizePKS . pubkeySize
     , stringrep = (key^.tkKey._1.pubkey & either (const "unknown") show . pubkeySize) ++ (key^.tkKey._1.pkalgo & pkalgoAbbrev)
   }
-  , keyUIDsAndUAts = FakeMap (map (second uidr) (processedOrOrig^.tkUIDs) ++ map (uatspsToString *** uidr) (processedOrOrig^.tkUAts))
+  , keyUIDsAndUAts = FakeMap (map (\(x,y) -> (x, uidr (Just x) y)) (processedOrOrig^.tkUIDs) ++ map (uatspsToString *** uidr Nothing) (processedOrOrig^.tkUAts))
   , keyBestOf = Nothing
 }
     where
@@ -177,21 +190,34 @@ checkKey mpt key = (\x -> x { keyBestOf = populateBestOf x }) KeyReport {
         uaspToString (OtherUASub t d) = "other-" ++ show t ++ ':':show (B.length d) ++ ':':BC8.unpack (Base16.encode (SHA3.hash 48 d))
         hdrToString (ImageHV1 JPEG) = "jpeg"
         hdrToString (ImageHV1 fmt) = "image-" ++ show (fromFVal fmt)
-        uidr sps = UIDReport (has sps) (map phas sps) (map (colorizeKETs (fromMaybe 0 mpt) (key^.tkKey._1.timestamp) . getKeyExpirationTimesFromSignature) sps) -- should that be 0?
-        populateBestOf krep = Just (UIDReport <$> best . uidSelfSigHashAlgorithms <*> best . uidPreferredHashAlgorithms <*> best . uidKeyExpirationTimes $ mconcat (justTheUIDRs krep))
-        justTheUIDRs = map snd . unFakeMap . keyUIDsAndUAts
+        uidr Nothing sps = Colored Nothing Nothing (UIDReport (has sps) (map phas sps) (map (colorizeKETs (fromMaybe 0 mpt) (key^.tkKey._1.timestamp) . getKeyExpirationTimesFromSignature) sps) (findRevocationReason sps)) -- should that be 0?
+        uidr (Just u) sps = colorizeUID u (UIDReport (has sps) (map phas sps) (map (colorizeKETs (fromMaybe 0 mpt) (key^.tkKey._1.timestamp) . getKeyExpirationTimesFromSignature) sps) (findRevocationReason sps)) -- should that be 0?
+        populateBestOf krep = Just (UIDReport <$> best . uidSelfSigHashAlgorithms <*> best . uidPreferredHashAlgorithms <*> best . uidKeyExpirationTimes <*> pure [] $ mconcat (justTheUIDRs krep))
+        justTheUIDRs = map (decolorize . snd) . unFakeMap . keyUIDsAndUAts
         best = take 1 . sortBy (comparing color)
+        decolorize (Colored _ _ x) = x
+        colorizeUID u
+            | '(' `elem` u = Colored (Just Yellow) (Just "parenthesis in uid")  -- FIXME: be more discerning
+            | not ('<' `elem` u) = Colored (Just Yellow) (Just "no left angle bracket in uid")  -- FIXME: be more discerning
+            | otherwise = Colored Nothing Nothing
+        findRevocationReason = concatMap grabReasons . filter isCertRevocationSig
+        isCertRevocationSig (SigV4 CertRevocationSig _ _ _ _ _ _) = True
+        isCertRevocationSig _ = False
+        grabReasons (SigV4 CertRevocationSig _ _ has _ _ _) = mapMaybe (grabReasons' . _sspPayload) has
+        grabReasons _ = []
+        grabReasons' (ReasonForRevocation a b) = Just (RevocationStatus True (show a) (decodeUtf8With lenientDecode b))
+        grabReasons' _ = Nothing
 
 prettyKeyReport :: POSIXTime -> TK -> Doc
 prettyKeyReport cpt key = do
  let keyReport = checkKey (Just cpt) key
  execWriter $ tell
    ( linebreak <> text "Key has potential validity:" <+> text (keyStatus keyReport)
-  <> linebreak <> text "Key has fingerprint:" <+> text (show (keyFingerprint keyReport))
+  <> linebreak <> text "Key has fingerprint:" <+> text (show (SpacedFingerprint (keyFingerprint keyReport)))
   <> linebreak <> text "Checking to see if key is OpenPGPv4:" <+> coloredToColor (text . show) (keyVer keyReport)
   <> linebreak <> (\kas -> text "Checking to see if key is RSA or DSA (>= 2048-bit):" <+> coloredToColor (text . show) (pubkeyalgo kas) <+> coloredToColor (text . maybe "unknown" show) (pubkeysize kas)) (keyAlgorithmAndSize keyReport)
   <> linebreak <> text "Checking user-ID- and user-attribute-related items:"
-  <> mconcat (map (uidtrip (keyCreationTime keyReport)) (unFakeMap (keyUIDsAndUAts keyReport)))
+  <> mconcat (map (uidtrip (keyCreationTime keyReport) . gottabeabetterway) (unFakeMap (keyUIDsAndUAts keyReport)))
   <> linebreak
    )
     where
@@ -199,11 +225,16 @@ prettyKeyReport cpt key = do
         coloredToColor f (Colored (Just Yellow) _ x) = yellow (f x)
         coloredToColor f (Colored (Just Red) _ x) = red (f x)
         coloredToColor f (Colored Nothing _ x) = f x
-        uidtrip ts (u, ur) = linebreak <> indent 2 (text u) <> colon
+        uidtrip ts (u, ur)
+           | null (uidRevocationStatus ur) = linebreak <> indent 2 (coloredToColor text u) <> colon
           <> linebreak <> indent 4 (text "Self-sig hash algorithms" <> colon <+> (list . map (coloredToColor (text . show)) . uidSelfSigHashAlgorithms) ur)
           <> linebreak <> indent 4 (text "Preferred hash algorithms" <> colon <+> mconcat (map ((linebreak <>) . indent 2 . coloredToColor (text . show)) (uidPreferredHashAlgorithms ur)))
           <> linebreak <> indent 4 (text "Key expiration times" <> colon <+> mconcat (map ((linebreak <>) . indent 2 . coloredToColor list . fmap (map (text . keyExp ts))) (uidKeyExpirationTimes ur)))
+           | otherwise = linebreak <> indent 2 (coloredToColor text u) <> colon <+> text "[revoked]"
+          <> linebreak <> indent 4 (text "Revocation code" <> colon <+> list (map (text . revocationCode) (uidRevocationStatus ur)))
+          <> linebreak <> indent 4 (text "Revocation reason" <> colon <+> list (map (text . T.unpack . revocationReason) (uidRevocationStatus ur)))
         keyExp ts ke = durationPrettyPrinter ke ++ " = " ++ formatTime defaultTimeLocale "%c" (posixSecondsToUTCTime (realToFrac ts + realToFrac ke))
+        gottabeabetterway (a, Colored x y z) = (Colored x y a, z)
 
 jsonReport :: POSIXTime -> TK -> BL.ByteString
 jsonReport ps = A.encode . checkKey (Just ps)
@@ -223,34 +254,41 @@ durationPrettyPrinter = concat . unfoldr durU
          | otherwise = Nothing
 
 data LintOutputFormat = Pretty | JSON | YAML
-    deriving (Eq, Read, Show)
+    deriving (Bounded, Enum, Eq, Read, Show)
 
 data LintOptions = LintOptions {
     lintOutputFormat :: LintOutputFormat
 }
 
-data Command = CmdLint LintOptions
+data Command = CmdLint LintOptions | CmdCanonicalize
 
 lintO :: Parser LintOptions
 lintO = LintOptions
-    <$> option
+    <$> option auto
         ( long "output-format"
        <> metavar "FORMAT"
        <> value Pretty
        <> showDefault
-       <> help "output format" )
+       <> ofHelp
+        )
+    where
+        ofHelp = helpDoc . Just $ text "output format" <> hardline <> list (map (text . show) ofchoices)
+        ofchoices = [minBound..maxBound] :: [LintOutputFormat]
+
 
 dispatch :: Command -> IO ()
 dispatch (CmdLint o) = banner' stderr >> hFlush stderr >> doLint o
+dispatch (CmdCanonicalize) = banner' stderr >> hFlush stderr >> doCanonicalize
 
 main :: IO ()
 main = do
     hSetBuffering stderr LineBuffering
-    customExecParser (prefs showHelpOnError) (info (helper <*> versioner <*> cmd) (header (banner "hokey") <> progDesc "hOpenPGP Key utility" <> footer (warranty "hokey"))) >>= dispatch
+    customExecParser (prefs showHelpOnError) (info (helper <*> versioner <*> cmd) (headerDoc (Just (banner "hokey")) <> progDesc "hOpenPGP Key utility" <> footerDoc (Just (warranty "hokey")))) >>= dispatch
 
 cmd :: Parser Command
 cmd = subparser
-    ( command "lint" (info ( CmdLint <$> lintO) ( progDesc "check key(s) for 'best practices'" ))
+    ( command "canonicalize" (info ( pure CmdCanonicalize) ( progDesc "arrange key components in a canonical ordering" ))
+   <> command "lint" (info ( CmdLint <$> lintO) ( progDesc "check key(s) for 'best practices'" ))
     )
 
 doLint :: LintOptions -> IO ()
@@ -263,8 +301,22 @@ doLint o = do
         output JSON cpt = mapM_ (BL.putStr . flip BL.append (BL.singleton 0x0a) . jsonReport cpt)
         output YAML cpt = mapM_ (B.putStr . yamlReport cpt)
 
+doCanonicalize :: IO ()
+doCanonicalize = do
+    runResourceT $ CB.sourceHandle stdin $= conduitGet get $=
+                   conduitToTKsDropping $= CL.map canonicalize $=
+                   conduitPut put $$ CB.sinkHandle stdout
+    where
+        canonicalize (TK k r ui ua s) = TK k
+                                           (sort r)
+                                           (indepthsort ui)
+                                           (indepthsort ua)
+                                           (indepthsort s)
+        indepthsort :: (Ord a, Ord b) => [(a,[b])] -> [(a,[b])]
+        indepthsort = nub . sort . over (mapped._2) sort
+
 banner' :: Handle -> IO ()
-banner' h = hPutStrLn h (banner "hokey" ++ "\n" ++ warranty "hokey")
+banner' h = hPutDoc h (banner "hokey" <> hardline <> warranty "hokey" <> hardline)
 
 sigissuer :: SignaturePayload -> Maybe EightOctetKeyId
 getIssuer :: SigSubPacketPayload -> Maybe EightOctetKeyId
